﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=7.31
@EndOfDesignText@
Sub Class_Globals
	Private xui As XUI
	Public cvs As B4XCanvas
	'input
	Type BCParagraphStyle (HorizontalAlignment As String, LineSpacingFactor As Float, MaxWidth As Int, Padding As B4XRect, WordWrap As Boolean)
	Type BCTextRun (TextFont As B4XFont, TextColor As Int, Text As String, CharacterSpacingFactor As Float, _
		VerticalOffset As Int, Underline As Boolean, BackgroundColor As Int, View As B4XView)
	'output
	Type BCParagraph (TextLines As List, CurrentLine As BCTextLine, Style As BCParagraphStyle, TwoLayers As Boolean, Width As Int, Height As Int, Views As List)
	Type BCTextLine (StartX As Int, BaselineY As Int, Height As Int, Unbreakables As List, Width As Int, EndsWithSoftLineBreak As Boolean, _
		MaxHeightAboveBaseLine As Int, ParentParagraph As BCParagraph, MaxHeightBelowBaseLine As Int)
	Type BCUnbreakableText (Width As Int, X As Int, Text As String, _
		IsMergable As Boolean, SingleStyleSections As List, ParentLine As BCTextLine)
	Type BCSingleStyleSection (StartX As Int, Glyphs As List, Run As BCTextRun, Width As Int, MaxHeightBelowBaseLine As Int, MaxHeightAboveBaseLine As Int, ParentUN As BCUnbreakableText)
	Type BCGlyph (cbc As CompressedBC, baseline As Int, Width As Int)
	Private CharBC As BitmapCreator
	Private cbccache As InternalCompressedBCCache
	Public mScale As Float = 1
	Private mSpaceBetweenCharacters As Float
	Private mSpaceBetweenLines As Int
	Private GlyphCache As Map
	Public DefaultColor As Int = xui.Color_Black
	Private WordBoundaries As String = "&*+-/().<>=\' ,:{}" & TAB & CRLF & Chr(13)
	Private ForegroundBC, BackgroundBC As BitmapCreator
	Private UnderlineWidth As Float = 1
	Private Brushes As Map
	Public DefaultStyle As BCParagraphStyle
	Public DefaultFont As B4XFont
	Private MinGapBetweenLines As Int = 5
	#if B4A
	Private stubForContext As Panel 'ignore
	#Else If B4J
	Private WritableImage As JavaObject
	Private sp As JavaObject
	#End If
End Sub

Public Sub Initialize
	Dim p As B4XView = xui.CreatePanel("")
	p.SetLayoutAnimated(0, 0, 0, 2dip, 2dip)
	#if B4i
	mScale = GetDeviceLayoutValues.NonnormalizedScale
	#Else If B4J
	Try
		Dim fx As JFX
		Dim jo As JavaObject = fx.PrimaryScreen
		mScale = 1 + Floor(jo.RunMethod("getOutputScaleX", Null))
	Catch
		Log(LastException)
		mScale = 1
	End Try
	Log("BCText scale: " & mScale)
	#end if
	setSpaceBetweenCharacters(100dip / 100)
	setSpaceBetweenLines(20dip)
	cvs.Initialize(p)
	ResizeCharBC(50dip * mScale, 50dip * mScale)
	Brushes.Initialize
	ResizeLayers(200dip, 100dip)
	cbccache.Initialize
	cbccache.ColorsMap.Initialize
	GlyphCache.Initialize
	Dim b(CharBC.SAME_COLOR_LENGTH_FOR_CACHE * 4 * CharBC.MAX_SAME_COLOR_SIZE + 4) As Byte
	cbccache.mBuffer = b
	DefaultFont = xui.CreateDefaultFont(16)
	DefaultStyle = CreateStyle

End Sub

Private Sub ResizeLayers (Width As Int, Height As Int)
	Width = Width * mScale
	Height = Height * mScale
	If ForegroundBC.IsInitialized = False Or Width > ForegroundBC.mWidth Or Height > ForegroundBC.mHeight Then
		Brushes.Clear
		ForegroundBC.Initialize(Width, Height)
		BackgroundBC.Initialize(ForegroundBC.mWidth, ForegroundBC.mHeight)
	Else
		ForegroundBC.DrawRect2(ForegroundBC.TargetRect, GetBrush(xui.Color_Transparent), True, 0)
		BackgroundBC.DrawRect2(BackgroundBC.TargetRect, GetBrush(xui.Color_Transparent), True, 0)
	End If
End Sub

Public Sub CreateStyle As BCParagraphStyle
	Dim s As BCParagraphStyle
	s.Initialize
	s.LineSpacingFactor = 1
	s.HorizontalAlignment = "Left"
	s.MaxWidth = 300dip
	s.Padding.Initialize(5dip, 5dip, 5dip, 5dip)
	s.WordWrap = True
	Return s
End Sub

Public Sub CreateRun (Text As String) As BCTextRun
	Dim r As BCTextRun
	r.Initialize
	r.BackgroundColor = 0
	r.CharacterSpacingFactor = 1
	r.TextFont = DefaultFont
	r.Text = Text
	r.TextColor = DefaultColor
	Return r
End Sub

Public Sub GenerateView (Runs As List, Style As BCParagraphStyle) As B4XView
	'layout
	Dim par As BCParagraph
	par.Initialize
	par.TextLines.Initialize
	par.Style = Style
	Dim unbreakeables As List
	unbreakeables.Initialize
	For Each run As BCTextRun In Runs
		HandleTextRun(run, unbreakeables)
	Next
	CreateLine(par)
	OrganizeUnbreakables(par, unbreakeables)
	OrganizeLines(par)
	'draw
	ResizeLayers(par.Width, par.Height)
	DrawParagraph(par, 0, 0)
	Dim pnl As B4XView = xui.CreatePanel("")
	pnl.SetLayoutAnimated(0, 0, 0, par.Width / mScale + Style.Padding.Left + Style.Padding.Right, _
		par.Height / mScale + Style.Padding.Top + Style.Padding.Bottom)
	If par.Width > 0 And par.Height > 0 Then
		If par.TwoLayers Then
			CreateImageView(BackgroundBC, par, pnl)
		End If
		CreateImageView(ForegroundBC, par, pnl)
	End If
	If par.Views.IsInitialized Then
		For Each v As B4XView In par.Views
			pnl.AddView(v, par.Style.Padding.Left + v.Left, par.Style.Padding.Top + v.Top, v.Width, v.Height)
		Next
	End If
	pnl.Tag = par
	Return pnl
End Sub

Private Sub CreateImageView (bc As BitmapCreator, par As BCParagraph, parent As B4XView)
	Dim iv As ImageView
	iv.Initialize("")
	Dim bmp As B4XBitmap = bc.Bitmap
	bc.SetBitmapToImageView(bmp.Crop(0, 0, par.Width, par.Height), iv)
	parent.AddView(iv, par.Style.Padding.Left, par.Style.Padding.Top, par.Width / mScale, par.Height / mScale)
End Sub


Private Sub OrganizeLines (p As BCParagraph)
	Dim alignment As String = p.Style.HorizontalAlignment.ToLowerCase
	Dim count As Int
	For Each line As BCTextLine In p.TextLines
		p.Width = Max(p.Width, line.Width)
		For Each un As BCUnbreakableText In line.Unbreakables
			For Each single As BCSingleStyleSection In un.SingleStyleSections
				line.MaxHeightAboveBaseLine = Max(single.MaxHeightAboveBaseLine, line.MaxHeightAboveBaseLine)
				line.MaxHeightBelowBaseLine = Max(single.MaxHeightBelowBaseLine, line.MaxHeightBelowBaseLine)
			Next
		Next
		If count = 0 Then
			line.Height = line.MaxHeightAboveBaseLine
		Else
			line.Height =  Max(line.MaxHeightAboveBaseLine + MinGapBetweenLines * mScale, mSpaceBetweenLines * p.Style.LineSpacingFactor)
		End If
		p.Height = p.Height + line.Height
		count = count + 1
	Next
	p.Width = Min((p.Style.MaxWidth - p.Style.Padding.Left - p.Style.Padding.Right) * mScale, p.Width)
	p.Height = p.Height + line.MaxHeightBelowBaseLine + 2 * mScale
	
	For Each line As BCTextLine In p.TextLines
		Select alignment
			Case "center"
				line.StartX = p.Width / 2 - line.Width / 2
			Case "right"
				line.StartX = p.Width - line.Width
			Case "justify"
				If line.EndsWithSoftLineBreak Then
					Dim last As BCUnbreakableText = line.Unbreakables.Get(line.Unbreakables.Size - 1)
					If IsSpace(last.Text) Then
						line.Unbreakables.RemoveAt(line.Unbreakables.Size - 1)
						line.Width = line.Width - last.Width
					End If
					Dim NumberOfGaps As Int = line.Unbreakables.Size - 1
					If NumberOfGaps > 0 Then
						Dim delta As Int = (p.Width - line.Width) / NumberOfGaps
						Dim accumalated As Int = 0
						For Each un As BCUnbreakableText In line.Unbreakables
							un.X = un.X + accumalated
							accumalated = accumalated + delta
						Next
					End If
				End If
		End Select
	Next
End Sub

Private Sub CreateLine(p As BCParagraph)
	Dim line As BCTextLine
	line.Initialize
	line.Unbreakables.Initialize
	line.ParentParagraph = p
	p.TextLines.Add(line)
	p.CurrentLine = line
End Sub


Private Sub HandleTextRun (Run As BCTextRun, Unbreakables As List)
	Dim DefaultFontCache As Map = GetFontCache(Run.TextFont, DefaultColor, Null)
	Dim FontCache As Map = GetFontCache(Run.TextFont, Run.TextColor, DefaultFontCache)
	Dim i1 As Int
	For i = 0 To Run.Text.Length - 1
		Dim c As String = Run.Text.CharAt(i)
		If WordBoundaries.Contains(c) Then
			If i >= i1 + 1 Then Unbreakables.Add(CreateUnbreakable(Run, Run.Text.SubString2(i1, i), FontCache, DefaultFontCache, False))
			Unbreakables.Add(CreateUnbreakable(Run, Run.Text.CharAt(i), FontCache, DefaultFontCache, True))
			i1 = i + 1
		Else If c = Chr(13) Then
			Continue
		End If
	Next
	If i1 < Run.Text.Length Then Unbreakables.Add(CreateUnbreakable(Run, Run.Text.SubString(i1), FontCache, DefaultFontCache, False))
	If Run.View.IsInitialized Then
		Unbreakables.Add(CreateUnbreakable(Run, "", FontCache, DefaultFontCache, False))
	End If
End Sub

Private Sub CreateUnbreakable (Run As BCTextRun, Text As String, FontCache As Map, DefaultFontCache As Map, IsSeparator As Boolean) As BCUnbreakableText
	Dim unbreakable As BCUnbreakableText
	unbreakable.Initialize
	unbreakable.SingleStyleSections.Initialize
	unbreakable.Text = Text
	unbreakable.IsMergable = Not(IsSeparator)
	Dim single As BCSingleStyleSection = CreateSingleSection(Run, Text, FontCache, DefaultFontCache)
	single.ParentUN = unbreakable
	unbreakable.SingleStyleSections.Add(single)
	unbreakable.Width = single.Width
	Return unbreakable
End Sub

Private Sub OrganizeUnbreakables (p As BCParagraph, unbreakables As List)
	If unbreakables.Size = 0 Then Return
	Dim un As BCUnbreakableText = unbreakables.Get(0)
	Dim i As Int = 1
	Do While i < unbreakables.Size
		Dim NextUn As BCUnbreakableText = unbreakables.Get(i)
		If un.IsMergable = True And NextUn.IsMergable = True Then
			MergeUnbreakables(un, NextUn)
			unbreakables.RemoveAt(i)
			i = i - 1
		Else
			un = NextUn
		End If
		i = i + 1
	Loop
	For Each un As BCUnbreakableText In unbreakables
		If un.Text = Chr(13) Then Continue
		If un.Text = Chr(10) Then
			CreateLine(p)
			Continue
		End If
		If p.CurrentLine.Unbreakables.Size > 0 And p.Style.WordWrap And _
				p.CurrentLine.Width + mSpaceBetweenCharacters + un.Width > (p.Style.MaxWidth - p.Style.Padding.Left - p.Style.Padding.Right) * mScale Then
			p.CurrentLine.EndsWithSoftLineBreak = True
			CreateLine(p)
		End If
		p.CurrentLine.Unbreakables.Add(un)
		un.ParentLine = p.CurrentLine
		If IsSpace(un.Text) And p.CurrentLine.Unbreakables.Size = 1 Then
			un.Width = 0
		End If
		If p.CurrentLine.Unbreakables.Size > 0 Then p.CurrentLine.Width = p.CurrentLine.Width + mSpaceBetweenCharacters
		un.X = p.CurrentLine.Width
		p.CurrentLine.Width = p.CurrentLine.Width + un.Width
	Next
End Sub

Private Sub MergeUnbreakables (un1 As BCUnbreakableText, un2 As BCUnbreakableText)
	un1.Width = un1.Width + un2.Width + mSpaceBetweenCharacters
	For Each single As BCSingleStyleSection In un2.SingleStyleSections
		single.ParentUN = un1
	Next
	un1.SingleStyleSections.AddAll(un2.SingleStyleSections)
	un1.Text = un1.Text & un2.Text
End Sub



Private Sub CreateSingleSection (Run As BCTextRun, Text As String, FontCache As Map, DefaultFontCache As Map) As BCSingleStyleSection
	Dim single As BCSingleStyleSection
	single.Initialize
	single.Glyphs.Initialize
	single.Run = Run
	For i = 0 To Text.Length - 1
		Dim g As BCGlyph = CreateGlyph(Text.CharAt(i), Run.TextFont, FontCache, DefaultFontCache, Run.TextColor)
		#if B4A AND DEBUG
		single.Glyphs.Add(Null)
		#End If
		single.Glyphs.Add(g)
		If g.cbc.IsInitialized Then
			single.MaxHeightAboveBaseLine = Max(single.MaxHeightAboveBaseLine, g.baseline + Run.VerticalOffset * mScale)
			single.MaxHeightBelowBaseLine = Max(single.MaxHeightBelowBaseLine, g.cbc.mHeight - g.baseline - Run.VerticalOffset * mScale)
		End If
		single.Width = single.Width + g.Width
		If i > 0 Then single.Width = single.Width + mSpaceBetweenCharacters * Run.CharacterSpacingFactor
	Next
	If Text = "" Then
		single.Width = Run.View.Width * mScale + mSpaceBetweenCharacters * 2 'extra spaces
		single.MaxHeightAboveBaseLine = (Run.View.Height + Run.VerticalOffset) * mScale
		single.MaxHeightBelowBaseLine = Run.View.Height * mScale - single.MaxHeightAboveBaseLine
	End If
	Return single
End Sub

Private Sub DrawParagraph (Paragraph As BCParagraph, X As Int, Y As Int)
	For Each line As BCTextLine In Paragraph.TextLines
		Y = Y + line.Height
		DrawLine(line, X, Y)
	Next
End Sub

Private Sub DrawLine(line As BCTextLine, OffsetX As Int, OffsetY As Int)
	line.BaseLineY = OffsetY
	For Each un As BCUnbreakableText In line.Unbreakables
		DrawUnbreakable(un, OffsetX + line.StartX, OffsetY)
	Next
End Sub

Private Sub DrawUnbreakable (un As BCUnbreakableText, OffsetX As Int, OffsetY As Int)
	Dim x As Int = OffsetX + un.X
	For Each single As BCSingleStyleSection In un.SingleStyleSections
		DrawSingleStyleSection(single, x, OffsetY)
		x = x + single.Width + mSpaceBetweenCharacters
	Next
End Sub

Private Sub DrawSingleStyleSection (single As BCSingleStyleSection, OffsetX As Int, OffsetY As Int)
	single.StartX = OffsetX
	For Each g As BCGlyph In single.Glyphs
		#if B4A and DEBUG

		If g = Null Then Continue
		#End If
		If g.cbc.IsInitialized Then
			ForegroundBC.DrawCompressedBitmap(g.cbc, g.cbc.TargetRect, OffsetX, OffsetY - g.baseline - single.Run.VerticalOffset * mScale)
		End If
		If single.Run.Underline Then
			ForegroundBC.DrawLine2(OffsetX, OffsetY + mScale, OffsetX + g.Width + mSpaceBetweenCharacters * single.Run.CharacterSpacingFactor, OffsetY + mScale, _
				GetBrush(single.Run.TextColor), UnderlineWidth * mScale)
		End If
		If single.Run.BackgroundColor <> 0 Then
			Dim line As BCTextLine = single.ParentUN.ParentLine
			Dim r As B4XRect
			r.Initialize(OffsetX, OffsetY - line.MaxHeightAboveBaseLine - 2 * mScale, _
				OffsetX + g.Width + mSpaceBetweenCharacters * single.Run.CharacterSpacingFactor, OffsetY + line.MaxHeightBelowBaseLine + 2 * mScale)
			BackgroundBC.DrawRect2(r, GetBrush(single.Run.BackgroundColor), True, 0)
			single.ParentUN.ParentLine.ParentParagraph.TwoLayers = True
		End If
		OffsetX = OffsetX + g.Width + mSpaceBetweenCharacters * single.Run.CharacterSpacingFactor
	Next
	If single.Glyphs.Size = 0 Then
		Dim v As B4XView = single.Run.View
		Dim par As BCParagraph = single.ParentUN.ParentLine.ParentParagraph
		If par.Views.IsInitialized = False Then par.Views.Initialize
		v.Left = (OffsetX + mSpaceBetweenCharacters) / mScale
		v.Top = OffsetY / mScale - v.Height - single.Run.VerticalOffset * mScale
		par.Views.Add(v)
	End If
End Sub

Private Sub ResizeCharBC(width As Int, height As Int)
	Dim ScaledWidth As Int = (width + 5) / mScale
	Dim ScaledHeight As Int = (height + 5) / mScale
	CharBC.Initialize(ScaledWidth * mScale, ScaledHeight * mScale)
	CharBC.MAX_SAME_COLOR_SIZE = 0
#if B4J
	CharBC.AlphaThresholdForCBCExtraction = 0
#End If
	cvs.Resize(ScaledWidth, ScaledHeight)
	#if b4J
	Dim w As Int = mScale * cvs.TargetRect.Width
	Dim h As Int = mScale * cvs.TargetRect.Height
	WritableImage.InitializeNewInstance("javafx.scene.image.WritableImage", Array(w, h))
	Dim sp As JavaObject
	sp.InitializeNewInstance("javafx.scene.SnapshotParameters", Null)
	Dim fx As JFX
	sp.RunMethod("setFill", Array(fx.Colors.Transparent))
	Dim transform As JavaObject
	transform.InitializeStatic("javafx.scene.transform.Transform")
	Dim scale As Double = mScale
	sp.RunMethod("setTransform", Array(transform.RunMethod("scale", Array(scale, scale))))
	#End If
End Sub

Public Sub FindSingleStyleSection (Paragraph As BCParagraph, X As Int, Y As Int) As BCSingleStyleSection
	x = (x - Paragraph.Style.Padding.Left) * mScale
	y = (y - Paragraph.Style.Padding.Top) * mScale
	For Each line As BCTextLine In Paragraph.TextLines
		If line.BaseLineY - line.MaxHeightAboveBaseLine <= y And line.BaseLineY + line.MaxHeightBelowBaseLine >= y Then
			For Each un As BCUnbreakableText In line.Unbreakables
				If un.X <= x And un.X + un.Width >= x Then
					For Each s As BCSingleStyleSection In un.SingleStyleSections
						If s.StartX <= x And s.StartX + s.Width >= x Then Return s
					Next
				End If
			Next
		End If
	Next
	Return Null
End Sub

Private Sub IsSpace(s As String) As Boolean
	Return s = " "
End Sub

Private Sub GetBrush(clr As Int) As BCBrush
	If Brushes.ContainsKey(clr) Then Return  Brushes.Get(clr)
	Dim b As BCBrush = ForegroundBC.CreateBrushFromColor(clr)
	Brushes.Put(clr, b)
	Return b
End Sub


Private Sub GetFontCache(Fnt As B4XFont, clr As Int, DefaultColorCache As Map) As Map
	Dim key As String = FontToKey(Fnt, clr)
	If GlyphCache.ContainsKey(key) Then Return GlyphCache.Get(key)
	Dim m As Map
	m.Initialize
	CreateGlyph("x", Fnt, m, DefaultColorCache, clr)
	GlyphCache.Put(key, m)
	Return m
End Sub

Private Sub CreateGlyph (c As String, Fnt As B4XFont, FontCache As Map, FontCacheDefaultColor As Map, clr As Int) As BCGlyph
	Dim g As BCGlyph = FontCache.Get(c)
	If g <> Null Then Return g
	If clr <> DefaultColor Then
		g = CreateGlyphFromDefaultColor(c, Fnt, FontCacheDefaultColor, clr)
	Else
		cvs.ClearRect(cvs.TargetRect)
		Dim r As B4XRect = cvs.MeasureText(c, Fnt)
		Dim BaseLine As Int = -r.Top + 5
		r.Left = r.Left * mScale
		r.Top = r.Top * mScale
		r.Right = r.Right * mScale
		r.Bottom = r.Bottom * mScale
		If CharBC.mWidth < r.Width + 20 * mScale Or CharBC.mHeight < r.Height + 20 * mScale Then
			ResizeCharBC(r.Width + 30 * mScale, r.Height + 30 * mScale)
		End If
'	cvs.DrawRect(cvs.TargetRect, xui.Color_Yellow, True, 0)
		cvs.DrawText(c, 5, BaseLine, Fnt, clr, "LEFT")
		#if B4A		
		Dim bmp As B4XBitmap = cvs.CreateBitmap
		#else if B4J
		Dim jo As JavaObject = cvs
		Dim bmp As B4XBitmap = jo.GetFieldJO("cvs").RunMethodJO("getObject", Null).RunMethod("snapshot", Array(sp, WritableImage))
		#else if B4i
		Dim cg As NativeObject = bmp
		Dim uiimg As NativeObject
		bmp = uiimg.Initialize("UIImage").RunMethod("imageWithCGImage:scale:orientation:", Array(cg.RunMethod("CGImage", Null), 1, _
			cg.GetField("imageOrientation")))
		#End If
		CharBC.CopyPixelsFromBitmap(bmp)
		Dim r2 As B4XRect = FindMinRect(r.Width + 20 * mScale, r.Height + 20 * mScale)
		Dim g As BCGlyph
		g.Initialize
		#if B4i
		g.baseline = Floor(-r.Top)
		#else
		g.baseline = BaseLine * mScale - r2.Top
		#End If
		If r2.Width > 0 Then
			r2.Left = Floor(r2.Left)
			r2.Right = Ceil(r2.Right)
			Dim cbc As CompressedBC = CharBC.ExtractCompressedBC(r2, cbccache)
			g.cbc = cbc
			g.width = cbc.mWidth
		Else
			Dim x As BCGlyph = FontCache.Get("x")
			If c = TAB Then
				g.Width = x.Width * 4
			Else
				g.width = x.width - 4 * mSpaceBetweenCharacters
			End If
			
		End If
	End If
	FontCache.Put(c, g)
	Return g
End Sub

Private Sub CreateGlyphFromDefaultColor(c As String, Fnt As B4XFont, DefaultColorCache As Map, clr As Int) As BCGlyph
	Dim BlackGlyph As BCGlyph = CreateGlyph(c, Fnt, DefaultColorCache, DefaultColorCache, DefaultColor)
	If BlackGlyph.cbc.IsInitialized = False Then Return BlackGlyph
	Dim g As BCGlyph
	g.Initialize
	g.baseline = BlackGlyph.baseline
	g.width = BlackGlyph.width
	g.cbc.Initialize
	g.cbc.Cache = BlackGlyph.cbc.Cache
	g.cbc.mHeight = BlackGlyph.cbc.mHeight
	g.cbc.mWidth = BlackGlyph.cbc.mWidth
	g.cbc.Rows = BlackGlyph.cbc.Rows
	g.cbc.TargetRect = BlackGlyph.cbc.TargetRect
	Dim buffer(BlackGlyph.cbc.mBuffer.Length) As Byte
	Dim argb As ARGBColor
	CharBC.ColorToARGB(clr, argb)
	Dim ai, ri, gi, bi As Int
	#if B4A or B4i
	ai = 3
	ri = 0
	gi = 1
	bi = 2
	#else if B4J
	ai = 3
	ri = 2
	gi = 1
	bi = 0
	#end if
	For i = 0 To buffer.Length - 1 Step 4
		#if B4i
		Dim a As Int = Bit.FastArrayGetByte(BlackGlyph.cbc.mBuffer, i + ai)
		Dim af As Float = a / 255
		Bit.FastArraySetByte(buffer, i + ri, argb.r * af)
		Bit.FastArraySetByte(buffer, i + gi, argb.g * af)
		Bit.FastArraySetByte(buffer, i + bi, argb.b * af)
		Bit.FastArraySetByte(buffer, i + ai, a)
#Else
		Dim a As Int = Bit.And(0xff, BlackGlyph.cbc.mBuffer(i + ai))
		Dim af As Float = a / 255
		buffer(i + ai) = a
		buffer(i + ri) = argb.r * af
		buffer(i + gi) = argb.g * af
		buffer(i + bi) = argb.b * af
#End If
	Next
	g.cbc.mBuffer = buffer
	Return g
End Sub


Private Sub FindMinRect (width As Int, height As Int) As B4XRect
	Dim r As B4XRect
	r.Initialize(width / 2, -1, -1, 0)
	For y = 0 To height - 1
		For x = 0 To width - 1
			If CharBC.IsTransparent(x, y) = False Then
				r.Left = Min(r.Left, x)
				Exit
			End If
		Next
		If x < width Then
			If r.Top = -1 Then
				r.Top = y
			Else
				r.Bottom = y + 1
			End If
			For x = width - 1 To 0 Step -1
				If CharBC.IsTransparent(x, y) = False Then
					r.Right = Max(r.Right, x + 1)
					Exit
				End If
			Next
		End If
	Next
	r.Bottom = Max(r.Bottom, r.Top + 1)
	Return r
End Sub

Private Sub FontToKey (fnt As B4XFont, Clr As Int) As String
	#if B4J or B4A
	Dim jo As JavaObject = fnt.ToNativeFont
	Return Clr + jo.RunMethod("hashCode", Null) + fnt.Size
	#Else
	Dim no As NativeObject = fnt.ToNativeFont
	Return Clr + no.GetField("hash").AsNumber
	#End If
End Sub

Public Sub getSpaceBetweenCharacters As Float
	Return mSpaceBetweenCharacters / mScale
End Sub

Public Sub setSpaceBetweenCharacters(f As Float)
	mSpaceBetweenCharacters = f * mScale
End Sub

Public Sub getSpaceBetweenLines As Float
	Return mSpaceBetweenLines / mScale
End Sub

Public Sub setSpaceBetweenLines(f As Float)
	mSpaceBetweenLines = f * mScale
End Sub

